/* tag: hosted forth environment, executable code
 *
 * Copyright (C) 2003 Patrick Mauritz, Stefan Reinauer
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/stat.h>

#ifdef __GLIBC__
#define _GNU_SOURCE
#include <getopt.h>
#endif

#include "config.h"
#include "types.h"
#include "dict.h"
#define HOSTED
#include "stack.h"
#include "plugins.h"

#define VERSION "1.1"

#define MEMORY_SIZE 1048576	/* 1M ram for hosted system */
#define DICTIONARY_SIZE 65536	/* 64k for the dictionary   */

/* prototypes */
void enterforth(ucell);
void printk(const char *fmt, ...);
static void exit_terminal(void);

extern unsigned char *dict;
extern cell dicthead;

extern ucell *last;

/* local variables */

ucell *latest, *state;
ucell *memory;

static int errors = 0;
static int segfault = 0;
int verbose = 0;

/* include path handling */
typedef struct include_path include;
struct include_path {
	char *path;
	include *next;
};

include includes = { ".", NULL };

/* the word names are used to generate the prim words in the
 * dictionary. This is done by the C written interpreter.
 */

static const char *wordnames[] = {
	"(semis)", "", "(lit)", "", "", "", "", "(do)", "(?do)", "(loop)",
	"(+loop)", "dup", "2dup", "?dup", "over", "2over", "pick", "drop",
	"2drop", "nip", "roll", "rot", "-rot", "swap", "2swap", ">r", "r>",
	"r@", "depth", "depth!", "rdepth", "rdepth!", "+", "-", "*", "u*",
	"mu/mod", "abs", "negate", "max", "min", "lshift", "rshift", ">>a",
	"and", "or", "xor", "invert", "d+", "d-", "m*", "um*", "@", "c@",
	"w@", "l@", "!", "+!", "c!", "w!", "l!", "=", ">", "<", "u>", "u<",
	"sp@", "move", "fill", "(emit)", "(?key)", "(key)", "execute",
	"here", "here!", "dobranch", "do?branch", "unaligned-w@",
	"unaligned-w!", "unaligned-l@", "unaligned-l!", "ioc@", "iow@",
	"iol@", "ioc!", "iow!", "iol!", "i", "j"
};

/*
 * skipws skips all whitespaces (space, tab, newline) from the input file
 */

static void skipws(FILE * f)
{
	int c;
	while (!feof(f)) {
		c = getc(f);

		if (c == ' ' || c == '\t' || c == '\n')
			continue;

		ungetc(c, f);
		break;
	}
}

/*
 * parse gets the next word from the input stream, delimited by
 * delim. If delim is 0, any word delimiter will end the stream
 * word delimiters are space, tab and newline. The resulting word
 * will be put zero delimited to the char array line.
 */

static int parse(FILE * f, char *line, char delim)
{
	int cnt = 0, c;

	while (!feof(f)) {
		c = getc(f);

		if (delim && c == delim)
			break;

		if ((!delim) && (c == ' ' || c == '\n' || c == '\t'))
			break;

		line[cnt++] = c;
	}

	line[cnt] = 0;

	return cnt;
}

/*
 * parse_word is a small helper that skips whitespaces before a word.
 * it's behaviour is similar to the forth version parse-word. 
 */

static void parse_word(FILE * f, char *line)
{
	skipws(f);
	parse(f, line, 0);
}


static void writestring(const char *str)
{
	unsigned int i;
	for (i = 0; i < strlen(str); i++) {
		dict[dicthead + i] = str[i];
	}
	dicthead += i + 1;
	dict[dicthead - 1] = (char) strlen(str) + 128;
}

#define writebyte(value) {*(dict+dicthead++)=value;}
#define writecell(value) {*(cell *)(dict+dicthead)=value; dicthead+=sizeof(cell);}
static void reveal(void)
{
	*last = *latest;
}


static void paddict(ucell align)
{
	while (dicthead % align != 0)
		writebyte(0);
}

static void write_dictionary(char *filename)
{
	FILE *f;
	ucell initxt;

	initxt = findword("initialize");
	if (initxt == (ucell) (-1))
		printk ("warning: dictionary needs word called initialize\n");

	f = fopen(filename, "w");
	if (!f) {
		printk("panic: can't open dictionary.\n");
		exit_terminal();
		exit(1);
	}

	fwrite(DICTID, 16, 1, f);
	fwrite(dict, dicthead, 1, f);

	/* Write start address and last to relocate on load */
	fwrite(&dict, sizeof(ucell), 1, f);
	fwrite(&last, sizeof(ucell), 1, f);

	fclose(f);

#ifdef DEBUG_DICTIONARY
	printk("wrote dictionary to file %s.\n", filename);
#endif
}

static ucell read_dictionary(char *fil)
{
	int ilen;
	ucell ret;
	char *mem;
	FILE *f;
	struct stat finfo;

	if (stat(fil, &finfo))
		return 0;

	ilen = finfo.st_size;

	if ((mem = malloc(ilen)) == NULL) {
		printk("panic: not enough memory.\n");
		exit_terminal();
		exit(1);
	}

	f = fopen(fil, "r");
	if (!f) {
		printk("panic: can't open dictionary.\n");
		exit_terminal();
		exit(1);
	}

	if (fread(mem, ilen, 1, f) != 1) {
		printk("panic: can't read dictionary.\n");
		fclose(f);
		exit_terminal();
		exit(1);
	}
	fclose(f);

	ret = load_dictionary(mem, ilen);

	free(mem);
	return ret;
}

static void fcreate(char *word, ucell cfaval)
{
	if (strlen(word) != 0) {
		writestring(word);
		/* get us at least 1 byte for flags */
		writebyte(0);
		paddict(sizeof(cell));
		/* set flags high bit. */
		dict[dicthead - 1] = 128;
		/* lfa and cfa */
		writecell(*latest);
		*latest = (ucell) dict + dicthead - sizeof(cell);
		writecell(cfaval);
	}
}


static ucell *buildvariable(char *name, cell defval)
{
	fcreate(name, DOVAR);	/* see dict.h for DOVAR and other CFA ids */
	writecell(defval);
	return (ucell *) (dict + dicthead - sizeof(cell));
}

static void buildconstant(char *name, cell defval)
{
	fcreate(name, DOCON);	/* see dict.h for DOCON and other CFA ids */
	writecell(defval);
}

static void builddefer(char *name)
{
	fcreate(name, DODFR);	/* see dict.h for DODFR and other CFA ids */
	writecell(0);
	writecell(findword("(semis)"));
}

static cell getvariable(char *name)
{
	cell cfa = findword(name);
#ifdef DEBUG_INTERPRETER
	printk("getting variable %s, cfa=%x\n", name, cfa);
#endif
	return *(cell *) (cfa + sizeof(cell));
}

static void add_includepath(char *path)
{
	include *incl = &includes;
	include *newpath;

	while (incl->next)
		incl = incl->next;

	newpath = malloc(sizeof(include));
	if (!newpath) {
		printk("panic: not enough memory for include path.\n");
		exit(1);
	}

	incl->next = newpath;
	newpath->path = path;
	newpath->next = NULL;
}

FILE *fopen_include(char *fil)
{
#define MAX_PATH_LEN 256
	char fullpath[MAX_PATH_LEN];
	FILE *ret;
	include *incl = &includes;

	while (incl) {
		strncpy(fullpath, incl->path, MAX_PATH_LEN);
		strncat(fullpath, "/", MAX_PATH_LEN);
		strncat(fullpath, fil, MAX_PATH_LEN);
		ret = fopen(fullpath, "r");
		if (ret != NULL)
			return ret;
		incl = incl->next;
	}
	return NULL;
}

static int interpret_source(char *fil)
{
	FILE *f;
	char tib[160];
	long num;
	char *test;

	const ucell SEMIS = findword("(semis)");
	const ucell LIT = findword("(lit)");
	const ucell DOBRANCH = findword("dobranch");

	if ((f = fopen_include(fil)) == NULL) {
		printk("error while loading source file '%s'\n", fil);
		errors++;
		return 1;
	}

	/* FIXME: We should read this file at 
	 * once. No need to get it char by char 
	 */

	while (!feof(f)) {
		ucell res;

		parse_word(f, tib);

		/* if there is actually no word, we continue right away */
		if (strlen(tib) == 0) {
			continue;
		}

		/* Checking for builtin words that are needed to
		 * bootstrap the forth base dictionary.
		 */

		if (!strcmp(tib, "(")) {
			parse(f, tib, ')');
			continue;
		}

		if (!strcmp(tib, "\\")) {
			parse(f, tib, '\n');
			continue;
		}

		if (!strcmp(tib, ":")) {
			parse_word(f, tib);

#ifdef DEBUG_INTERPRETER
			printk("create colon word %s\n\n", tib);
#endif
			fcreate(tib, DOCOL);	/* see dict.h for DOCOL and other CFA ids */
			*state = (ucell) (-1);
			continue;
		}

		if (!strcmp(tib, ";")) {
#ifdef DEBUG_INTERPRETER
			printk("finish colon definition\n\n");
#endif
			writecell(SEMIS);
			*state = (ucell) 0;
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "variable")) {
			parse_word(f, tib);
#ifdef DEBUG_INTERPRETER
			printk("defining variable %s\n\n", tib);
#endif
			buildvariable(tib, 0);
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "constant")) {
			parse_word(f, tib);
#ifdef DEBUG_INTERPRETER
			printk("defining constant %s\n\n", tib);
#endif
			buildconstant(tib, POP());
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "value")) {
			parse_word(f, tib);
#ifdef DEBUG_INTERPRETER
			printk("defining value %s\n\n", tib);
#endif
			buildconstant(tib, POP());
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "defer")) {
			parse_word(f, tib);
#ifdef DEBUG_INTERPRETER
			printk("defining defer word %s\n\n", tib);
#endif
			builddefer(tib);
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "include")) {
			parse_word(f, tib);
#ifdef DEBUG_INTERPRETER
			printk("including file %s\n\n", tib);
#endif
			interpret_source(tib);
			continue;
		}

		if (!strcmp(tib, "[']")) {
			ucell xt;
			parse_word(f, tib);
			xt = findword(tib);
			if (*state == 0) {
#ifdef DEBUG_INTERPRETER
				printk
				    ("writing address of %s to stack\n\n",
				     tib);
#endif
				PUSH(xt);
			} else {
#ifdef DEBUG_INTERPRETER
				printk("writing lit, addr(%s) to dict\n\n",
				       tib);
#endif
				writecell(LIT);	/* lit */
				writecell(xt);
			}
			continue;
			/* we have no error detection here */
		}

		if (!strcasecmp(tib, "s\"")) {
			int cnt;
			cell loco;

			cnt = parse(f, tib, '"');
#ifdef DEBUG_INTERPRETER
			printk("compiling string %s\n", tib);
#endif
			loco = dicthead + (6 * sizeof(cell));
			writecell(LIT);
			writecell((ucell) dict + loco);
			writecell(LIT);
			writecell(cnt);
			writecell(DOBRANCH);
			loco = cnt + sizeof(cell) - 1;
			loco &= ~(sizeof(cell) - 1);
			writecell(loco);
			memcpy(dict + dicthead, tib, cnt);
			dicthead += cnt;
			paddict(sizeof(cell));
			continue;
		}

		/* look if tib is in dictionary. */
		/* should the dictionary be searched before the builtins ? */
		res = findword(tib);
		if (res != (ucell) (-1)) {
			u8 flags = *(u8 *) (res - sizeof(cell) - 1);
#ifdef DEBUG_INTERPRETER
			printk("%s is 0x%x\n", tib, (ucell) res);
#endif
			if (!(*state) || (flags & 3)) {
#ifdef DEBUG_INTERPRETER
				printk("executing %s, %x (flags: %s %s)\n",
				       tib, res,
				       (flags & 1) ? "immediate" : "",
				       (flags & 2) ? "compile-only" : "");
#endif
				PC = res;
				enterforth(res);
			} else {
#ifdef DEBUG_INTERPRETER
				printk("writing %s to dict\n\n", tib);
#endif
				writecell(res);
			}
			continue;
		}


		/* if not look if it's a number */
		if (tib[0] == '-')
			num = strtol(tib, &test, getvariable("base"));
		else
			num = strtoul(tib, &test, getvariable("base"));

		if (*test != 0) {
			/* what is it?? */
			printk("%s is not defined.\n\n", tib);
			errors++;
#ifdef DEBUG_INTERPRETER
			continue;
#else
			return -1;
#endif
		}

		if (*state == 0) {
#ifdef DEBUG_INTERPRETER
			printk("pushed %x to stack\n\n", num);
#endif
			PUSH(num);
		} else {
#ifdef DEBUG_INTERPRETER
			printk("writing lit, %x to dict\n\n", num);
#endif
			writecell(LIT);	/* lit */
			writecell(num);
		}
	}
	fclose(f);

	return 0;
}

static int build_dictionary(void)
{
	ucell lfa = 0;
	unsigned int i;

	/* we need a temporary place for latest 
	 * outside the dictionary
	 */

	latest = &lfa;

#ifdef DEBUG_DICTIONARY
	printk("building dictionary, %d primitives.\nbuilt words:",
	       sizeof(wordnames) / sizeof(void *));
#endif

	for (i = 0; i < sizeof(wordnames) / sizeof(void *); i++) {
		if (strlen(wordnames[i]) != 0) {
			fcreate((char *) wordnames[i], i);
#ifdef DEBUG_DICTIONARY
			printk(" %s", wordnames[i]);
#endif
		}
	}
#ifdef DEBUG_DICTIONARY
	printk(".\n");
#endif

	/* get last/latest and state */
	state = buildvariable("state", 0);
	last = buildvariable("forth-last", 0);
	latest = buildvariable("latest", 0);

	*latest = (ucell) (latest - 2);

	buildvariable("base", 10);
	buildconstant("/c", sizeof(uint8_t));
	buildconstant("/w", sizeof(uint16_t));
	buildconstant("/l", sizeof(uint32_t));
	buildconstant("/n", sizeof(void *));

	reveal();
	printk("finished building dictionary.\n");

	return 0;
}

/*
 * functions used by primitives
 */

int availchar(void)
{
	int tmp = getc(stdin);
	if (tmp != EOF) {
		ungetc(tmp, stdin);
		return -1;
	}
	return 0;
}

u8 inb(u32 reg)
{
	io_ops_t *ior=find_iorange(reg);
	if (ior)
		return ior->inb(reg);

	printk("TRAP: io byte read @0x%x", reg);
	return 0xff;
}

u16 inw(u32 reg)
{
	io_ops_t *ior=find_iorange(reg);
	if (ior)
		return ior->inw(reg);

	printk("TRAP: io word read @0x%x", reg);
	return 0xffff;
}

u32 inl(u32 reg)
{
	io_ops_t *ior=find_iorange(reg);
	if (ior)
		return ior->inl(reg);

	printk("TRAP: io long read @0x%x", reg);
	return 0xffffffff;
}

void outb(u32 reg, u8 val)
{
	io_ops_t *ior=find_iorange(reg);
	if (ior) {
		ior->outb(reg, val);
		return;
	}

	printk("TRAP: io byte write 0x%x -> 0x%x", val, reg);
}

void outw(u32 reg, u16 val)
{
	io_ops_t *ior=find_iorange(reg);
	if (ior) {
		ior->outw(reg, val);
		return;
	}

	printk("TRAP: io word write 0x%x -> 0x%x", val, reg);
}

void outl(u32 reg, u32 val)
{
	io_ops_t *ior=find_iorange(reg);
	if (ior) {
		ior->outl(reg, val);
		return;
	}

	printk("TRAP: io long write 0x%x -> 0x%x", val, reg);
}

/*
 * terminal initialization and cleanup.
 */

static void init_terminal(void)
{
	struct termios termios;

	tcgetattr(0, &termios);
	termios.c_lflag &= ~(ICANON | ECHO);
	tcsetattr(0, 0, &termios);
}

static void exit_terminal(void)
{
	struct termios termios;

	tcgetattr(0, &termios);
	termios.c_lflag |= (ICANON | ECHO);
	tcsetattr(0, 0, &termios);
}

/*
 *  segmentation fault handler. linux specific?
 */

static void
segv_handler(int signo __attribute__ ((unused)),
	     siginfo_t * si, void *context __attribute__ ((unused)))
{
	static int count = 0;
	ucell addr = 0xdeadbeef;

	if (count) {
		printk("Died while dumping forth dictionary core.\n");
		goto out;
	}

	count++;

	if (PC >= (ucell) dict && PC <= (ucell) dict + dicthead)
		addr = *(ucell *) PC;

	printk("panic: segmentation violation at %x\n", si->si_addr);
	printk("dict=0x%x here=0x%x(dict+0x%x) pc=0x%x(dict+0x%x)\n",
	       dict, dict + dicthead, dicthead, PC, PC - (ucell) dict);
	printk("dstackcnt=%d rstackcnt=%d instruction=%x\n",
	       dstackcnt, rstackcnt, addr);

#ifdef DEBUG_DSTACK
	printdstack();
#endif
#ifdef DEBUG_RSTACK
	printrstack();
#endif
	printk("Writing dictionary core file\n");
	write_dictionary("forth.dict.core");

      out:
	exit_terminal();
	exit(1);
}

/* 
 * allocate memory and prepare engine for memory management.
 */

void init_memory(void)
{
	memory = malloc(MEMORY_SIZE);
	if (!memory) {
		printk("panic: not enough memory on host system.\n");
		exit_terminal();
		exit(1);
	}

	/* we push start and end of memory to the stack
	 * so that it can be used by the forth word QUIT
	 * to initialize the memory allocator
	 */

	PUSH((ucell) memory);
	PUSH((ucell) memory + MEMORY_SIZE);
}

/*
 * main loop
 */

#define BANNER	"OpenBIOS core. (C) 2003 Patrick Mauritz, Stefan Reinauer\n"\
		"This software comes with absolutely no warranty. "\
		"All rights reserved.\n\n"


#define USAGE   "usage: %s [options] [dictionary file|source file]\n\n"

int main(int argc, char *argv[])
{
	struct sigaction sa;
	unsigned char *dictname = NULL;
	int c;

	const char *optstring = "VhsI:D:P:p:?";

	printk(BANNER);

	while (1) {
#ifdef __GLIBC__
		int option_index = 0;
		static struct option long_options[] = {
			{"version", 0, NULL, 'V'},
			{"verbose", 0, NULL, 'v'},
			{"help", 0, NULL, 'h'},
			{"include", 1, NULL, 'I'},
			{"dictionary", 1, NULL, 'D'},
			{"segfault", 0, NULL, 's'},
			{"plugin-path", 1, NULL, 'P'},
			{"plugin", 1, NULL, 'p'}
		};

		c = getopt_long(argc, argv, optstring, long_options,
				&option_index);
#else
		c = getopt(argc, argv, optstring);
#endif
		if (c == -1)
			break;

		switch (c) {
		case 'V':
			printk("Version " VERSION "\n");
			return 0;
		case 'h':
		case '?':
			printk("Version " VERSION "\n" USAGE, argv[0]);
			return 0;
		case 'v':
			verbose = 1;
			break;
		case 's':
			segfault = 1;
			break;
		case 'I':
			printk("adding '%s' to include path\n", optarg);
			add_includepath(optarg);
			break;
		case 'D':
			printk("Dumping final dictionary to '%s'\n", optarg);
			dictname = optarg;
			break;
		case 'P':
			printk("Plugin search path is now '%s'\n", optarg);
			plugindir=optarg;
			break;
		case 'p':
			printk("Loading plugin %s\n", optarg);
			load_plugin(optarg);
			break;
		default:
			return 1;
		}
	}

	if (argc < optind + 1) {
		printk(USAGE, argv[0]);
		return 1;
	}

	if ((dict = (unsigned char *) malloc(DICTIONARY_SIZE)) == NULL) {
		printk("panic: not enough memory.\n");
		return 1;
	}

	if (!segfault) {
		if (verbose)
			printk("Installing SIGSEGV handler...");

		sa.sa_sigaction = segv_handler;
		sigemptyset(&sa.sa_mask);
		sa.sa_flags = SA_SIGINFO | SA_NODEFER;
		sigaction(SIGSEGV, &sa, 0);

		if (verbose)
			printk("done.\n");
	}

	/* set terminal to do non blocking reads */
	init_terminal();
	read_dictionary(argv[optind]);
	PC = findword("initialize");
	if (PC != (ucell) (-1)) {
		if (verbose) {
			if (optind + 1 != argc)
				printk("Warning: only first dictionary used.\n");

			printk("dictionary loaded (%d bytes).\n", dicthead);
			printk("Initializing memory...");
		}
		init_memory();

		if (verbose) {
			printk("done\n");

			printk("Jumping to dictionary...");
		}

		enterforth(PC);

		if (dictname != NULL)
			write_dictionary(dictname);

		free(memory);

	} else {		/* input file is not a dictionary */

		build_dictionary();

		while (optind < argc)
			interpret_source(argv[optind++]);

		printk("interpretion finished. %d errors occured.\n",
		       errors);

#ifndef DEBUG_INTERPRETER
		if (errors)
			printk("dictionary not dumped to file.\n");
		else
#endif
		{
			if (dictname)
				write_dictionary(dictname);
			else
				write_dictionary("bootstrap.dict");
		}
	}

	exit_terminal();
	free(dict);
	return 0;
}
